#!/bin/bash

set -m
set -e

DEBUG=${DEBUG:-0}

log() {
  local heading="$1"
  shift
  local args="$@"
  printf "  \033[36m%10s\033[0m : \033[90m%s\033[0m\n" "$heading" "$args" >&2
}

abort() {
  local args="$@"
  printf "\n  \033[31mERR!: $args\033[0m\n\n" >&2 && exit 1
}

debug() {
  if [[ $DEBUG -eq 1 ]]; then
    log "$@"
  fi
  return 0
}

BASE_URL="${BASE_URL:-http://localhost:8091}"
MEMORY_QUOTA=2258
INDEX_MEMORY_QUOTA=512
FTS_MEMORY_QUOTA=301
SERVICES="index,kv,n1ql"

CB_INDEX_PATH="/opt/couchbase/var/lib/couchbase/data"
CB_PATH="$CB_INDEX_PATH"

CURL_OPTS=(
  "-L"
)

BUCKETS=(
  "widgets"
)

INDEXES=(
  "create index w_org_id_idx on \`widgets\`(organizationId);"
  "create index w_id_idx on \`widgets\`(id);"
)

CURL() {
  debug curl "$@"
  curl -f $@
}

post() {
  local url="$@"
  CURL -X POST "${CURL_OPTS[@]}" $url
}

get() {
  local url="$@"
  CURL "${CURL_OPTS[@]}" $url
}

fetch_initial() {
  log fetch initial settings...
  local url="$BASE_URL/nodes/self"
  get -s "$url"
}

update_cluster_settings() {
  log update cluster settings...
  local url="$BASE_URL/nodes/self/controller/settings"
  post -s "$url"                       \
    -d "index_path=$CB_INDEX_PATH"     \
    -d "path=$CB_PATH"
}

rename() {
  local url="$BASE_URL/node/controller/rename"
  post -s "$url" -d "hostname=127.0.0.1"
}

setup_services() {
  local url="$BASE_URL/node/controller/setupServices"
  post -s "$url" -d "services=kv,index,n1ql"
}

setup_web_interface() {
  local url="$BASE_URL/settings/web"
  post -s "$url"            \
    -d "password=password"  \
    -d "port=SAME"          \
    -d "username=username"
}

setup_indexes() {
  local url="$BASE_URL/settings/indexes"
  post -s "$url" -d "storageMode=forestdb"
}

get_indexes() {
  local url="$BASE_URL/settings/indexes"
  local out=$(get -s "$url" | jq .storageMode)
  local res=$(echo "$out" | sed 's/"//g')
  if [[ $res -ne "forestdb" ]]; then
    abort "invalid storageMode for indexes\n$out"
  fi
}

create_bucket() {
  local url="$BASE_URL/pools/default/buckets"
  local name="$1"
  local ram="${2:-300}"

  post -s "$url"                     \
    -d "authType=sasl"               \
    -d "autoCompactionDefined=false" \
    -d "bucketType=membase"          \
    -d "evictionPolicy=valueOnly"    \
    -d "flushEnabled=0"              \
    -d "name=$name"                  \
    -d "ramQuotaMB=$ram"             \
    -d "replicaIndex=0"              \
    -d "replicaNumber=0"             \
    -d "saslPassword="               \
    -d "threadNumber=3"
}

setup_cluster() {
  local SETTINGS=$(fetch_initial)
  local MEMORY_QUOTA=$(echo "$SETTINGS" | jq ".memoryQuota")
  local INDEX_MEMORY_QUOTA=$(echo "$SETTINGS" | jq ".indexMemoryQuota")
  local FTS_MEMORY_QUOTA=$(echo "$SETTINGS" | jq ".ftsMemoryQuota")

  log found "  -> memoryQuota: $MEMORY_QUOTA"
  log found "  -> indexMemoryQuota: $INDEX_MEMORY_QUOTA"
  log found "  -> ftsMemoryQuota: $FTS_MEMORY_QUOTA"

  update_cluster_settings || abort unable to update cluster settings

  log "rename" cluster...
  rename || abort unable to rename

  log setup services...
  setup_services || abort unable to setup services

  log setup web interface...
  setup_web_interface || abort unable to setup web interface

  CURL_OPTS+=("-u username:password")

  echo
  log setup indexes...
  setup_indexes

  echo
  log check index storage...
  get_indexes
}

print_exit() {
  local str="$1"
  local out=$(echo "$str" | jq .status 2>/dev/null)
  local s=$(echo "$out" | sed 's/"//g')
  if [ "$s" = "success" ]; then
    return 0
  else
    return 1
  fi
}

run_full() {
  local query="$@"
  actually_run_query "$query" | sed 's/cbq> //'
}

run_query() {
  local query="$@"
  # try it twice
  log cbq "$query" attempt 0
  local res=$(run_full "$query")
  if ! print_exit "$res"; then
    sleep 5
    log cbq "$query" attempt 1
    res=$(run_full "$query")
    if ! print_exit "$res"; then
      abort "$res"
    fi
  fi
}

actually_run_query() {
  local q="$1"
  echo "$q" | cbq -c "username:password" -e http://localhost:8093 --quiet=true
}

create_primary_index() {
  local bucket="$1"
  run_query "CREATE PRIMARY INDEX ON $bucket USING GSI;"
}

main() {
  log start server
  /entrypoint.sh couchbase-server &
  log waiting server spinup

  sleep 15
  log setup cluster
  setup_cluster
  log created cluster
  echo
  log buckets create

  for bucket in ${BUCKETS[*]}; do
    log create bucket $bucket
    create_bucket "$bucket"
  done

  log created buckets

  log waiting bucket initialization
  sleep 10
  echo

  log indexes create
  for bucket in ${BUCKETS[*]}; do
    log index primary index for bucket: $bucket
    create_primary_index $bucket
  done

  for index in "${INDEXES[@]}"; do
    log index "$index"
    run_query "$index"
  done
  log created indexes
}

main

# fg 1
