'use strict'

const EE = require('events')
const couchbase = require('couchbase')
const Async = require('async')
const Query = couchbase.N1qlQuery
const ERRORS = require('./errors')
const noop = () => {}

const defaultLog = {
  silly: noop
  , verbose: noop
  , info: noop
  , http: noop
  , warn: noop
  , error: noop
}

function normalizeError(err) {
  if (!err) return err
  err.code = ERRORS.byNumber(err.code) || err.code
  return err
}

module.exports = class DB extends EE {
  constructor(opts) {
    super()

    opts = Object.assign({
      uri: 'couchbase://localhost'
      , connectionTimeout: 5000
      , operationTimeout: 2500
      , retryMaxAttempts: 20
      , retryMaxDelay: 60000
      , retryBackoff: 5000
    }, opts)

    this.uri = opts.uri
    this.cluster = null
    this.buckets = new Map()
    this.log = opts.log
      ? opts.log.child('db')
      : defaultLog

    this.connected = false
    this._connecting = false

    this.connectionAttempts = 0
    this.connectionTimeout = opts.connectionTimeout
    this.operationTimeout = opts.operationTimeout
    this.retryMaxAttempts = opts.retryMaxAttempts
    this.retryMaxDelay = opts.retryMaxDelay
    this.retryBackoff = opts.retryBackoff
    this._createClient(opts.mock)
  }

  _createClient(mock) {
    /* istanbul ignore next */
    const Cluster = mock
      ? couchbase.Mock.Cluster
      : couchbase.Cluster

    this.cluster = new Cluster(this.uri)
  }

  query(type, sql, params, cb) {
    if (!this._checkBucket(type, cb)) return
    const bucket = this.buckets.get(type)

    const query = Query.fromString(sql)
    query
    // prepare the statement
      .adhoc(false)
      // strong consistency per request
      .consistency(Query.Consistency.REQUEST_PLUS)

    bucket.query(query, params, (err, result, meta) => {
      if (err) return cb(normalizeError(err))
      cb(null, result, meta)
    })
  }

  get(bucketName, key, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.get(key, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  getAndLock(bucketName, key, opts, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.getAndLock(key, opts, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  getMulti(bucketName, keys, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.getMulti(keys, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  remove(bucketName, key, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.remove(key, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  upsert(bucketName, key, val, opts, cb) {
    if (typeof opts === 'function') {
      cb = opts
      opts = {}
    }
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.upsert(key, val, opts, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  insert(bucketName, key, val, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.insert(key, val, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  replace(bucketName, key, val, opts, cb) {
    if (typeof opts === 'function') {
      cb = opts
      opts = {}
    }
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.replace(key, val, opts, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  counter(bucketName, key, delta, opts, cb) {
    if (typeof opts === 'function') {
      cb = opts
      opts = {}
    }
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.counter(key, delta, opts, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  unlock(bucketName, key, cas, cb) {
    if (!this._checkBucket(bucketName, cb)) return
    const bucket = this.buckets.get(bucketName)
    bucket.unlock(key, cas, (err, res) => {
      if (err) return cb(normalizeError(err))
      cb(null, res)
    })
  }

  _checkBucket(bucketName, cb) {
    if (typeof cb !== 'function') {
      throw new TypeError('cb must be a function')
    }
    if (!this.buckets.has(bucketName)) {
      // bucket not open
      const err = new Error(`Bucket not found: ${bucketName}`)
      // bucketNotFound error
      err.code = 25
      normalizeError(err)
      setImmediate(() => {
        cb(err)
      })
      return false
    }

    return true
  }

  disconnect() {
    this._connecting = false
    this.connected = false
    // TODO(evanlucas) Update couchbase (https://github.com/couchbase/couchnode)
    // to support callbacks for the disconnect event on a bucket
    for (const bucket of this.buckets.values()) {
      bucket.disconnect()
    }
  }

  connect(buckets, cb) {
    this._connecting = true
    Async.each(buckets, (bucket, cb) => {
      this.openBucket(bucket, '', cb)
    }, (err) => {
      this._connecting = false
      if (err) {
        // Retry on network errors
        if (err.code === 'networkError') {
          return this._reconnect(buckets, cb)
        }

        this.log.error('Unable to connect', err)
        normalizeError(err)
        return cb && cb(err)
      }

      this.connected = true
      this.connectionAttempts = 0
      this.emit('connect')
      return cb && cb()
    })
  }

  _reconnect(buckets, cb) {
    this.connectionAttempts += 1

    this.log.info('retrying connection to couchbase')

    if (this.connectionAttempts > this.retryMaxAttempts) {
      const err = new Error(this.retryMaxAttempts + 'connection failures')
      this.log.error('Maximum database connection attempts exhausted', err)
      return cb && cb(err)
    }

    const delay = this._getConnectionDelay(
      this.connectionAttempts
      , this.retryBackoff
      , this.retryMaxDelay
    )
    setTimeout(() => { this.connect(buckets, cb) }, delay)
  }

  // Returns delay period before next connection attempt (in msecs)
  _getConnectionDelay(attempts, backoff, maxDelay) {
    if (attempts * backoff < maxDelay) {
      return attempts * backoff
    }

    return maxDelay
  }


  openBucket(name, password, cb) {
    const bucket = this.cluster.openBucket(name, password, (err) => {
      if (err) {
        this.log.error('unable to open bucket', err)
        normalizeError(err)
        return cb(err)
      }

      this.buckets.set(name, bucket)
      cb(null, bucket)
    })

    bucket.connectionTimeout = this.connectionTimeout
    bucket.operationTimeout = this.operationTimeout
    return bucket
  }
}

module.exports.ERRORS = ERRORS
